(*----------------------------------------------------------------------(C)-*)
(* Copyright (C) 2006-2017 Konstantin Korovin and The University of Manchester. 
   This file is part of iProver - a theorem prover for first-order logic.

   iProver is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 2 of the License, or 
   (at your option) any later version.
   iProver is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
   See the GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with iProver.  If not, see <http://www.gnu.org/licenses/>.         *)
(*----------------------------------------------------------------------[C]-*)

open Lib
open Options
open Statistics
open Logic_interface

(*----- debug modifiable part-----*)

let dbg_flag = false
   
type dbg_gr = 
  |D_trace
  |D_uc
  |D_uc_ref

let dbg_gr_to_str = function 
 |D_trace -> "trace"
 |D_uc -> "uc" 
 |D_uc_ref -> "uc_ref" 

let dbg_groups =
    [
     (* D_trace; *)
     D_uc;
     D_uc_ref;
   ]

let module_name = "abstr_ref_arg_filter"
  
(*----- debug fixed part --------*)
  
let () = dbg_flag_msg dbg_flag module_name
  
let dbg group str_lazy = 
    Lib.dbg_out_pref dbg_flag dbg_groups group dbg_gr_to_str module_name str_lazy
  
let dbg_env group f = 
    Lib.dbg_env_set dbg_flag dbg_groups group f
  
(*----- debug -----*)

(*-----  flags ------*)
let add_sf_axioms_flag = false  (* only needed for model represntation *)
let _ = out_warning ("add_sf_axioms_flag: "^(string_of_bool add_sf_axioms_flag))

    
(* continue with last abstraction for inter-schedule runs; can be moved to options *)
let glb_reuse_sig_abstr_flag = true

(* TODO: move to options! *)
  type refine_bsig_param =  ARS_Trivial | ARS_Single (* ARS_Trivial is concretising all symbols in unsat core; ARS_Single is flip first false arg *)

  let bsp_param_to_str bsp = 
    match bsp with 
    | ARS_Trivial -> "ARS_Trivial"
    | ARS_Single ->  "ARS_Single"
  
  let refine_bsig_param =  (*ARS_Trivial*)  ARS_Single 
  

  let _ = out_warning (bsp_param_to_str refine_bsig_param)

(*-----  end flags ------*)

let glb_swp_cnt = ref 0        (* counter should be global since after schedule restarts the counter should not be reset *)
let glb_abstr_loop_cnt = ref 0 (* counter for main eabstr loop iterations, used for adjoint predicates; as above this cnt should be global *)


(*
c-> c_signature
signature -> abstr signature
c_abstr \/ p_switch

refine:
c_abstr -> c 
*)

type bool_sig = bool list (* [b_1;..;b_n] if true then the argument present and is not otherwise *)

(* abstr_symb : asymb *)

type symb_filter = symbol * bool_sig (* symbol and signature containing both true and false *)

(*
type abstr_symb_wrap = (* abstract symb warp: asw *)
    {
     asw_symb_filter      : symb_filter;
     asw_abstr_symb       : symbol; (* abstr_symb*)
   }
*)

module SF_Key = 
  struct
  
   type t = symb_filter

   let compare (f1,bs1) (f2,bs2) = 
     if (f1 == f2)
     then 
       list_compare_lex Pervasives.compare bs1 bs2 
     else
       Symbol.compare f1 f2  
  end
   
module SF_Map = Map.Make(SF_Key)

module TLMap = Term.MapList
module TLSet = Term.SetList



(*--------------------------------*)
type sf_ax_wrap = 
    {
     mutable sf_axs : clause list;  (* axiom: f_0_1_2(x,y) <-> f(x,z,y) | ~ax_switch *)
     mutable sf_swp : term; (* ax_switch *)
   }


(*--------------------------------*)
type abstr_sig = (* asg; global abstr symbols *)
    {

(* sig fiter *)
     mutable asg_sf_to_as            : symbol SF_Map.t;    (* (f,bool_sig) -> abstr_symb *)
     mutable asg_as_to_sf            : symb_filter SMap.t; (* abstr_symb -> (f,bool_sig) *)

(* filter axioms *)
(*     mutable asg_sf_axioms           : CSet.t;  (* axiom: f_0_1_2(x,y) <-> f(x,z,y) | ~ax_switch; *) *)
     mutable asg_as_to_axw       : sf_ax_wrap SMap.t;   (* abstr_symb -> sf_ax_wrap *)
   }

let create_abstr_sig () = 
  {
   asg_sf_to_as      = SF_Map.empty;
   asg_as_to_sf      = SMap.empty;
(*   asg_sf_axioms     = CSet.empty; *)
   asg_as_to_axw = SMap.empty;
 }

(*------ reuse abstr_sig abster restarts --------*)

let glb_abstr_sig = create_abstr_sig ()

(*--------------------------------*)
let create_switch_pred cnt =
  let sw_pred_name = "$$iProver_ar_switch_pred_"^(string_of_int cnt) in
  add_pred_0 ~is_special:true sw_pred_name

let create_adjoint_pred cnt = 
  let adj_pred_name = "$$iProver_ar_adj_pred_"^(string_of_int cnt) in
  add_pred_0 ~is_special:true adj_pred_name


let get_new_sf_ax_swp () = (* just continue switch_preds *)
  let swp = create_switch_pred !glb_swp_cnt in
  glb_swp_cnt := !glb_swp_cnt +1;
  swp
 
(*--------------------------------*)

(* filters elements of list corresponding to b (true/false) positions in bsig *)
let bs_filter_list b bsig list = 
  assert ((List.length bsig) = (List.length list));
  let f res current_b e = 
    if (b = current_b)  then e::res else res 
  in
  List.rev (List.fold_left2 f [] bsig list)

(* list of b (true/false) positions numbering starts with 1 *)
let bs_filter_positions b bsig = 
  let f (cnt,res) e = if (e = b) then (cnt+1,cnt::res) else (cnt+1,res) in
  let (_cnt, res) = (List.fold_left f (1,[]) bsig) in
  List.rev res

let bs_and_2 bs1 bs2 = List.map2 (&&) bs1 bs2   

(* non-variable filter sig *)
let non_var_bsig tlist = List.map (fun t -> not (Term.is_var t)) tlist 

(*--------*)
let abstr_sig_add_abstr_symb abstr_sig orig_symb bsig = 
  dbg D_trace (lazy ("abstr_sig_add_abstr_symb: orig: "^(Symbol.to_string orig_symb)));   
  let symb_filter = (orig_symb, bsig) in
  try 
    let abstr_symb = SF_Map.find symb_filter abstr_sig.asg_sf_to_as in
    abstr_symb
  with 
    Not_found -> 
      let positions = bs_filter_positions true bsig in
 (* create abstr_symb *)
      let abstr_symb_name = (* TODO: add some $$ prefix ?*)
        String.concat "_"
          ((Symbol.to_string orig_symb)::
           ((string_of_int 0)::
            (List.map string_of_int positions))) 
      in 
      let (orig_arg_types, orig_val_type) = Symbol.get_stype_args_val_def orig_symb in
      let abstr_arg_types = bs_filter_list true bsig orig_arg_types in
      let abstr_symb_type = create_stype abstr_arg_types orig_val_type in 
      let abstr_symb = create_symbol abstr_symb_name abstr_symb_type in

      dbg D_trace (lazy ("env_add_abstr_symb: abstr: "^(Symbol.to_string abstr_symb)));   

(* add abstr_symb to env. *)    
      abstr_sig.asg_sf_to_as <- SF_Map.add symb_filter abstr_symb abstr_sig.asg_sf_to_as;
      abstr_sig.asg_as_to_sf <- SMap.add abstr_symb symb_filter abstr_sig.asg_as_to_sf;

(* add filter axioms *)
      let fresh_vars_env = Var.init_fresh_vars_env () in
      let tvar_list_orig = 
        let f rest arg_type = 
          (add_var_term (Var.get_next_fresh_var fresh_vars_env arg_type))::rest 
        in
        List.rev (List.fold_left f [] orig_arg_types) 
      in

      let orig_term = add_fun_term orig_symb tvar_list_orig in

      let tvar_list_abstr = bs_filter_list true bsig tvar_list_orig in  
      let abstr_term = add_fun_term abstr_symb tvar_list_abstr in
      let ax_switch_pred = get_new_sf_ax_swp () in
      let neg_ax_switch_pred = add_neg_atom ax_switch_pred in
      let new_filter_axioms =
        if (orig_val_type == Symbol.symb_bool_type) (* predicate axiom *)
        then        
            let axiom_orig_impl_abstr = 
              create_clause Clause.tstp_source_arg_filter_axiom [(add_neg_atom orig_term); abstr_term; neg_ax_switch_pred] 
            in
            let axiom_abstr_impl_orig = 
              create_clause Clause.tstp_source_arg_filter_axiom [(add_neg_atom abstr_term); orig_term; neg_ax_switch_pred] 
            in
            [axiom_orig_impl_abstr;axiom_abstr_impl_orig]

        else (* equational axiom *)
          let eq_atom = add_typed_equality_sym orig_val_type orig_term abstr_term in
          let axiom_eq_filter_axiom = 
            create_clause Clause.tstp_source_arg_filter_axiom [eq_atom;neg_ax_switch_pred]
          in
          [axiom_eq_filter_axiom]
      in
      dbg D_trace (lazy ("env_add_abstr_symb: eq_filter_axioms: "^(Clause.clause_list_to_string new_filter_axioms)));   
(*      ar_env.ar_current_assumptions <- TSet.add ax_switch_pred ar_env.ar_current_assumptions; *)
      let sf_ax_wrap = 
        {
         sf_axs = new_filter_axioms;
         sf_swp = ax_switch_pred;
       }
      in
      abstr_sig.asg_as_to_axw <- SMap.add abstr_symb sf_ax_wrap abstr_sig.asg_as_to_axw;
(*      abstr_sig.asg_sf_axioms <- List.fold_left (fun rest c ->  CSet.add c rest) abstr_sig.asg_sf_axioms new_filter_axioms; *)
      abstr_symb      

(* returns (orig_symb, bsig) or raises Not_found  *)
  let abstr_sig_get_orig_symb abstr_sig abstr_symb = 
    SMap.find abstr_symb abstr_sig.asg_as_to_sf 
 
(*------------------------*)
type ar_sig_env =  (* current abstr_sig env *)
    {

(* global *)
     mutable ase_abstr_sig         : abstr_sig; 

(* current ar *)
     mutable ase_symb_to_asymb     : symbol SMap.t;      (* symb -> abstr_symb *)
                                                         (* which are not in this map are assumed to be concretised *)   
     mutable ar_sf_axioms          : CSet.t; (* sf axioms corresponding to current abstr symbs *)
     mutable ar_sig_assumptions    : TSet.t; (* assumptions correpsonding to current sf axioms *)
   }

let create_ar_sig_env () = 
  {
   ase_abstr_sig = create_abstr_sig ();
   ase_symb_to_asymb = SMap.empty;
   
   ar_sf_axioms = CSet.empty;
   ar_sig_assumptions  = TSet.empty;
 }

(* global reuse of glb_abstr_sig *)
let create_ar_sig_env_glb () = 
  let ar_sig_env = create_ar_sig_env () in 
  ar_sig_env.ase_abstr_sig <- glb_abstr_sig;
  ar_sig_env

(* keep current sig abstraction globally *)
let glb_ar_sig_env = create_ar_sig_env_glb ()

(* remove abstraction of the orig_symb from the current abstraction *)
  let ar_sig_env_rm_abstr_symb ar_sig_env orig_symb =
    let abstr_sig = ar_sig_env.ase_abstr_sig in
    try 
      let abstr_symb = SMap.find orig_symb ar_sig_env.ase_symb_to_asymb in
      let sf_axw = SMap.find abstr_symb abstr_sig.asg_as_to_axw in
      ar_sig_env.ase_symb_to_asymb <- SMap.remove orig_symb ar_sig_env.ase_symb_to_asymb;
      ar_sig_env.ar_sf_axioms <- List.fold_left (fun rest c ->  CSet.remove c rest) ar_sig_env.ar_sf_axioms sf_axw.sf_axs;
      ar_sig_env.ar_sig_assumptions <- TSet.remove sf_axw.sf_swp ar_sig_env.ar_sig_assumptions; (* add complement ? if so then complement also need to be removed when adding *)
    with 
      Not_found -> ()

(* replaces the old abstr symb in the current abstr *)
  let ar_sig_env_replace_abstr_symb ar_sig_env orig_symb bsig =
    let abstr_sig = ar_sig_env.ase_abstr_sig in
    ar_sig_env_rm_abstr_symb ar_sig_env orig_symb; (* first remove current abstraction *)
    let abstr_symb = abstr_sig_add_abstr_symb abstr_sig orig_symb bsig in
    let sf_axw = SMap.find abstr_symb abstr_sig.asg_as_to_axw in
    ar_sig_env.ase_symb_to_asymb <- SMap.add orig_symb abstr_symb ar_sig_env.ase_symb_to_asymb;
    ar_sig_env.ar_sf_axioms <- List.fold_left (fun rest c ->  CSet.add c rest) ar_sig_env.ar_sf_axioms sf_axw.sf_axs;
    ar_sig_env.ar_sig_assumptions <- TSet.add sf_axw.sf_swp ar_sig_env.ar_sig_assumptions;
    abstr_symb
  
(* returns (abstr_symb, bsig) or raises Not_found  *)
  let get_abstr_symb ar_sig_env orig_symb = 
    let abstr_sig = ar_sig_env.ase_abstr_sig in
    let abstr_symb  = SMap.find orig_symb ar_sig_env.ase_symb_to_asymb in 
    let (_, bsig) = SMap.find abstr_symb abstr_sig.asg_as_to_sf in
    (abstr_symb, bsig)

  let ar_sig_env_all_concretised ar_sig_env = SMap.is_empty ar_sig_env.ase_symb_to_asymb

  let ar_sig_env_get_all_clauses ar_sig_env     = 
    dbg D_trace (lazy ("-----------------------"));
    dbg D_trace (lazy ("ar_sig_env_get_all_clauses: ar_sf_axioms: \n "^(Clause.clause_list_to_string (CSet.elements ar_sig_env.ar_sf_axioms))));   
    ar_sig_env.ar_sf_axioms
      
  let ar_sig_env_get_all_assumptions ar_sig_env = 
    dbg D_trace (lazy ("ar_sig_env_get_all_assumptions: sf axioms assumptions:"^(Term.term_list_to_string (TSet.elements ar_sig_env.ar_sig_assumptions))));   
    ar_sig_env.ar_sig_assumptions

(*--------------------------------*)

(* abstract clause: acl *)

type abstr_cl_wrap = (* aclw *)
    {
     aclw_orig_cl : clause;
     aclw_lits : lit list;    (* without the switch pred *)
     aclw_switch_pred : term; (* switch_pred -- prop pred *)
     aclw_acl : clause;       (* acl: aclw_lits \/ ~switch_pred *)
   }

type abstr_clauses = (* global sotrage *)
    {
(* abstr clauses *)
     mutable ar_swp_to_aclw         : abstr_cl_wrap TMap.t; (* switch_pred -> abstr_cl_wrap  *)
     mutable ar_abstr_lits_to_swp   : term TLMap.t;         (* lists of abstr literals -> switch preds *)     
   }

let create_abstr_clauses () =
  {
   ar_swp_to_aclw        = TMap.empty;
   ar_abstr_lits_to_swp  = TLMap.empty;
 }

(* global; reused after restarts *)
let glb_abstr_clauses = create_abstr_clauses ()

let get_cl_switch_pred abstr_clauses lit_list = 
  dbg D_trace (lazy ("get_cl_switch_pred: "^(Term.term_list_to_string lit_list)));   
  try 
    TLMap.find lit_list abstr_clauses.ar_abstr_lits_to_swp  
  with 
    Not_found ->
      ( 
        let sw_pred = create_switch_pred !glb_swp_cnt in
        dbg D_trace (lazy ("get_cl_switch_pred: Not_found: "^(Term.term_list_to_string lit_list)^" swp: "^(Term.to_string sw_pred)));   
        abstr_clauses.ar_abstr_lits_to_swp <- TLMap.add lit_list sw_pred abstr_clauses.ar_abstr_lits_to_swp;
        glb_swp_cnt := !glb_swp_cnt +1;
        sw_pred 
       )

(* lit_list wihtout switch pred; returns (abstr_clause, swp) *)
let abstr_clauses_add abstr_clauses orig_clause abstr_lits = 
  let switch_pred = get_cl_switch_pred abstr_clauses abstr_lits in (* normalise ? or is it ok because orig_lits are normalised ?*)
  try 
    let abstr_cl_warp = TMap.find switch_pred abstr_clauses.ar_swp_to_aclw in 
    (abstr_cl_warp.aclw_acl, switch_pred) 
  with 
    Not_found -> 
      begin         
        let compl_switch_pred = add_neg_atom switch_pred in
        
(* filtered clause *)
        let tstp_source = Clause.tstp_source_arg_filter_abstr orig_clause in
        let abstr_clause = create_clause tstp_source (compl_switch_pred::abstr_lits) in
        Clause.inherit_conjecture abstr_clause orig_clause;
        let abstr_cl_wrap = (* aclw *)
          {
           aclw_orig_cl = orig_clause;
           aclw_lits = abstr_lits;         (* without the switch pred *)
           aclw_switch_pred = switch_pred; (* switch_pred -- prop pred *)
           aclw_acl = abstr_clause;        (* acl: ~aclw_lits \/ switch_pred *)
         }
        in      
        dbg D_trace (lazy ("abstr_clauses_add abstr_clauses: abstr_clause: "^(Clause.to_string abstr_clause))); 
        
(*      ar_env.ar_acl_to_aclw <- CMap.add abstr_clause abstr_cl_wrap ar_env.ar_acl_to_aclw; *)
        abstr_clauses.ar_swp_to_aclw <- TMap.add switch_pred abstr_cl_wrap abstr_clauses.ar_swp_to_aclw;
        (abstr_clause, switch_pred)        
      end
(*---------------------------*)

type ar_cl_env = (* ar environment containing current abstr of clauses *)
    {
(* global *)
     mutable ar_abstr_clauses       : abstr_clauses;

(*  *)
     mutable ar_acl_set             : CSet.t;
     mutable ar_concretised_set     : CSet.t; 
     mutable ar_cl_assumptions      : TSet.t;
   }

let create_ar_cl_env () = 
  {
   ar_abstr_clauses = create_abstr_clauses ();
   ar_acl_set = CSet.empty;
   ar_concretised_set = CSet.empty;
   ar_cl_assumptions = TSet.empty;
 }  

let create_ar_cl_env_glb () = 
  let ar_cl_env = create_ar_cl_env () in 
  ar_cl_env.ar_abstr_clauses <- glb_abstr_clauses;
  ar_cl_env

let ar_cl_env_clear ar_cl_env = (* keep only global ar_abstr_clauses *)
  ar_cl_env.ar_acl_set <- CSet.empty;
  ar_cl_env.ar_concretised_set <- CSet.empty;
  ar_cl_env.ar_cl_assumptions <- TSet.empty


(* removes current abstraction of the clause for the current abstraction (but keeps in ar_abstr_clauses) *)
let ar_cl_env_rm_acl ar_cl_env swp = 
  let ar_abstr_clauses = ar_cl_env.ar_abstr_clauses in
  let abstr_cl_wrap = TMap.find swp ar_abstr_clauses.ar_swp_to_aclw in 
  ar_cl_env.ar_acl_set <- CSet.remove abstr_cl_wrap.aclw_orig_cl ar_cl_env.ar_acl_set;
  ar_cl_env.ar_cl_assumptions <- TSet.remove abstr_cl_wrap.aclw_switch_pred ar_cl_env.ar_cl_assumptions (* add compl pred ? then take care whan adding  abstr clauses*)

(* adds new abstr; old abstr is not removed ! use rm above *)
let ar_cl_env_add_acl ar_cl_env orig_clause abstr_lits = 
  let ar_abstr_clauses = ar_cl_env.ar_abstr_clauses in
  let (abstr_cl, switch_pred) = abstr_clauses_add ar_abstr_clauses orig_clause abstr_lits in
  ar_cl_env.ar_acl_set <- CSet.add abstr_cl ar_cl_env.ar_acl_set;
  ar_cl_env.ar_cl_assumptions <- TSet.add switch_pred ar_cl_env.ar_cl_assumptions

(* add/rm concrete clauses *)
let ar_cl_env_rm_conr_cl ar_cl_env concrete_cl =
  ar_cl_env.ar_concretised_set <- CSet.remove concrete_cl ar_cl_env.ar_concretised_set

let ar_cl_env_add_conr_cl ar_cl_env concrete_cl = 
  ar_cl_env.ar_concretised_set <- CSet.add concrete_cl ar_cl_env.ar_concretised_set

let ar_cl_env_get_all_clauses     ar_cl_env = 
  dbg D_trace (lazy ("-----------------------"));
  dbg D_trace (lazy ("ar_cl_env_get_all_clauses: abstr clauses: \n "^(Clause.clause_list_to_string (CSet.elements ar_cl_env.ar_acl_set))));   
  dbg D_trace (lazy ("-----------------------"));
  dbg D_trace (lazy ("ar_cl_env_get_all_clauses: concretised clauses: \n "^(Clause.clause_list_to_string (CSet.elements ar_cl_env.ar_concretised_set))));   
  CSet.union ar_cl_env.ar_acl_set ar_cl_env.ar_concretised_set

let ar_cl_env_get_all_assumptions ar_cl_env = 
  dbg D_trace (lazy ("ar_cl_env_get_all_assumptions: "^(Term.term_list_to_string (TSet.elements ar_cl_env.ar_cl_assumptions))));   
  ar_cl_env.ar_cl_assumptions

(*---------------------------*)
type ar_env = 
    {
     mutable ar_orig_clause_set     : CSet.t;

     mutable ar_sig_env             : ar_sig_env;
     mutable ar_cl_env              : ar_cl_env;

     mutable ar_adjoint_preds       : term list; (* current adjoint predicates *)
   }

let create_ar_env () = 
    {
     ar_orig_clause_set            = CSet.empty;

     ar_sig_env                    = create_ar_sig_env ();
     ar_cl_env                     = create_ar_cl_env ();

     ar_adjoint_preds              = []; 
   }

(* global reuse of ar_sig_env/ar_cl_env *)
let create_ar_env_glb () = 
  {
   ar_orig_clause_set            = CSet.empty;

   ar_sig_env                    = glb_ar_sig_env; (* create_ar_sig_env_glb (); *)
   ar_cl_env                     = create_ar_cl_env_glb ();

   ar_adjoint_preds              = []; 
 }





let ar_env_get_all_clauses ar_env = 
  if add_sf_axioms_flag 
  then
      CSet.union (ar_sig_env_get_all_clauses ar_env.ar_sig_env) (ar_cl_env_get_all_clauses ar_env.ar_cl_env)
  else
    ar_cl_env_get_all_clauses ar_env.ar_cl_env  

let ar_env_get_all_assumptions ar_env = 
  let sig_cl_ass = 
    if add_sf_axioms_flag 
    then
        TSet.union (ar_sig_env_get_all_assumptions ar_env.ar_sig_env) (ar_cl_env_get_all_assumptions ar_env.ar_cl_env) 
    else
       ar_cl_env_get_all_assumptions ar_env.ar_cl_env
  in 
  let all_ass = TSet.union (TSet.of_list ar_env.ar_adjoint_preds) sig_cl_ass in 
  all_ass

let ar_env_get_all_adjoint_preds ar_env = ar_env.ar_adjoint_preds


(* returns: symb_to_sf is a map symb -> currenft arg filter for this symbol *)        
  let rec ground_abstr_tsig symb_to_sf t = 
    if (Term.is_ground t) 
    then symb_to_sf
    else
      begin
        match t with
        |Term.Fun (symb, args, _info) ->
            let args_list = Term.arg_to_list args in 
            let non_var_bsig = non_var_bsig args_list in
            let non_var_args = bs_filter_list true non_var_bsig args_list in
            let exists_var_arg = (List.exists (fun e -> e = false) non_var_bsig) in
            let new_symb_to_sf = 
              if exists_var_arg 
              then
                try 
                  let old_sf = SMap.find symb symb_to_sf in 
                  let new_sf = bs_and_2 non_var_bsig old_sf in
                  SMap.add symb new_sf symb_to_sf 
                with 
                  Not_found ->
                    SMap.add symb non_var_bsig symb_to_sf
              else
                symb_to_sf
            in
            List.fold_left ground_abstr_tsig new_symb_to_sf non_var_args 
              
        |_-> failwith "ground_abstr_tsig: t should be a non-variable term"
      end
              

  let ground_abstr_clsig symb_to_sf c = List.fold_left ground_abstr_tsig symb_to_sf (Clause.get_literals c)
  
  let ground_abstr_cl_list_sig cl_list = List.fold_left ground_abstr_clsig SMap.empty cl_list
    

(*------------------------------*)
(* should be applied to atoms or terms (not to negations) *)
(* we assume that sf is already filled-in in the ar_sig_env *)

  let rec abstr_term ar_sig_env t = 
 (*   if (Term.is_ground t) 
    then t
    else
*)
      begin
        match t with
        |Term.Fun (orig_symb, args, _info) ->
          let args_list = Term.arg_to_list args in 
          let (new_symb, new_args) = 
            try 
              let (new_symb,bsig) = get_abstr_symb ar_sig_env orig_symb in (*can raise Not_found *)
              let filtered_args = bs_filter_list true bsig args_list in
              let new_args =  List.map (abstr_term ar_sig_env) filtered_args in 
              (new_symb,new_args)
            with
              Not_found -> 
                (
                 let new_args = List.map (abstr_term ar_sig_env) args_list in 
                 (orig_symb,new_args)
                )
          in          
          add_fun_term new_symb new_args
        | Term.Var _ -> t
      end


  type abstr_res =  Concretised of clause | Abstr_cl of clause * term (* clause switch pred *)

(* abstracts orig_clause *)
(* changes current clauses in ar_cl_env *)
  let apply_sig_abstr_to_clause ar_sig_env ar_cl_env orig_clause = 
    dbg D_trace (lazy ("abstr_clause: orig_clause: "^(Clause.to_string orig_clause))); 
    let orig_lits = Clause.get_literals orig_clause in 
    let ground_abstr_lit lit = add_term_db (Term.apply_to_atom (abstr_term ar_sig_env) lit) in 
    let abstr_lits = List.map ground_abstr_lit orig_lits in
    let non_trivial = List.exists2 (fun l1 l2 -> not (l1 == l2)) orig_lits abstr_lits in
    if non_trivial
    then 
      begin (* add abstr cl *)
        ar_cl_env_add_acl ar_cl_env orig_clause abstr_lits
      end
    else
      begin (* orig_clause = abstr *)
        ar_cl_env_add_conr_cl ar_cl_env orig_clause   
      end


(*---------*)
  let flip_first_false bsig = 
    let flipped_flag = ref false in 
    let f rest b = 
      if (!flipped_flag || b)
      then
        b::rest 
      else
        (
         flipped_flag:=true;
         true::rest
        )
    in
    let rev_flipped = List.fold_left f [] bsig in
    List.rev rev_flipped
      
(*---------------------*)      
  let refine_bsig bsig = 
    match refine_bsig_param with
    | ARS_Trivial -> List.map (fun _ -> true) bsig
    | ARS_Single  -> flip_first_false bsig

(* returns: set of orig symbols to refine *)
(*----v1: swps -> abstr_cl_wrap -> orig_clause -> orig_symb (-> ref_abstr_symb) *)
  let get_orig_sig_to_refine_v1 ar_env swps = 
    let ar_sig_env = ar_env.ar_sig_env in
    let ar_cl_env = ar_env.ar_cl_env in
    let ar_abstr_clauses = ar_cl_env.ar_abstr_clauses in
    let f symb_rest swp =
      try
        let abstr_cl_wrap = TMap.find swp ar_abstr_clauses.ar_swp_to_aclw in
        let orig_clause = abstr_cl_wrap.aclw_orig_cl in
        let is_relevant_symb symb = SMap.mem symb ar_sig_env.ase_symb_to_asymb in
        let cl_rel_symbs = Clause.find_all_sym ~is_relevant_symb orig_clause in
        SSet.union cl_rel_symbs symb_rest 
      with 
        Not_found -> symb_rest
    in
    List.fold_left f SSet.empty swps

(*----v2: swps -> abstr_cl_wrap -> abstr_symb -> orig_symb (-> ref_abstr_symb) *)
 let get_orig_sig_to_refine_v2 ar_env swps = 
    let ar_sig_env = ar_env.ar_sig_env in
    let ar_cl_env = ar_env.ar_cl_env in
    let ar_abstr_clauses = ar_cl_env.ar_abstr_clauses in
    let abstr_symbs_uc = 
      let f asymb_rest swp =
        try 
          let abstr_cl_wrap = TMap.find swp ar_abstr_clauses.ar_swp_to_aclw in
          let acl_asymbs =  Clause.find_all_sym ~is_relevant_symb:(fun _ -> true) abstr_cl_wrap.aclw_acl in
          SSet.union acl_asymbs asymb_rest 
       with 
          Not_found -> asymb_rest
      in
      List.fold_left f SSet.empty swps
    in
    let orig_symbs_uc = 
      let f abstr_symb symb_rest =        
        try
          let (orig_symb,_) = abstr_sig_get_orig_symb ar_sig_env.ase_abstr_sig abstr_symb in
          SSet.add orig_symb symb_rest
        with 
          Not_found -> 
            symb_rest
      in
      SSet.fold f abstr_symbs_uc SSet.empty
    in
    orig_symbs_uc

let get_orig_sig_to_refine = get_orig_sig_to_refine_v2

(* ----- *)
let refine_symb_ar_sig_env ar_sig_env orig_symb = 
  try 
    let (curr_abstr_symb, curr_bsig) = get_abstr_symb ar_sig_env orig_symb in
    let new_bsig = refine_bsig curr_bsig in 
    let non_trivial = List.exists (fun e -> e = false) new_bsig in
    if non_trivial 
    then
      begin
        let _new_abstr_symb = ar_sig_env_replace_abstr_symb ar_sig_env orig_symb new_bsig in ()
      end
    else
      begin
        ar_sig_env_rm_abstr_symb ar_sig_env orig_symb  (* concretise this symbol *)
      end
  with 
    Not_found -> ()

  let refine_symbs_ar_sig_env ar_sig_env orig_symbs = SSet.iter (refine_symb_ar_sig_env ar_sig_env) orig_symbs

  let refine_ar_env ar_env swps =  (* swps from uc *)
    let ar_sig_env = ar_env.ar_sig_env in
    let ar_cl_env = ar_env.ar_cl_env in
    let orig_symbs = get_orig_sig_to_refine ar_env swps in
    refine_symbs_ar_sig_env ar_sig_env orig_symbs;

(* TODO: clean! *)
    ar_cl_env_clear ar_cl_env;
    if (ar_sig_env_all_concretised ar_sig_env)
    then 
      (ar_cl_env.ar_concretised_set <- ar_env.ar_orig_clause_set)
    else
      (CSet.iter (apply_sig_abstr_to_clause ar_sig_env ar_cl_env) ar_env.ar_orig_clause_set)


(*-----------------------*)
  let get_orig_clauses_swps ar_env swps = 
    let ar_cl_env = ar_env.ar_cl_env in
    let ar_abstr_clauses = ar_cl_env.ar_abstr_clauses in
    let f cl_rest swp =
      try
        let abstr_cl_wrap = TMap.find swp ar_abstr_clauses.ar_swp_to_aclw in
        let orig_clause = abstr_cl_wrap.aclw_orig_cl in
        orig_clause::cl_rest
      with 
        Not_found -> cl_rest
    in
    List.fold_left f [] swps


(*----------------------*)
 
(* tech. ref.  *)
let init_abstr_ref = ref true

(*----------------------*)        
  let init_ar_env cl_list = 
(*    let ar_env = create_ar_env () in *)
    let ar_env = 
      if glb_reuse_sig_abstr_flag 
      then
          create_ar_env_glb () 
      else
        create_ar_env ()
    in
    let ar_sig_env = ar_env.ar_sig_env in
    let ar_cl_env = ar_env.ar_cl_env in
    
    ar_env.ar_orig_clause_set <- CSet.of_list cl_list;

(* fill signaure filter *)
    ( if !init_abstr_ref ||  (not glb_reuse_sig_abstr_flag)
    then 
      (let gr_abstr_sig_map = ground_abstr_cl_list_sig cl_list in
      SMap.iter (fun orig_symb bsig -> ignore (ar_sig_env_replace_abstr_symb ar_sig_env orig_symb bsig)) gr_abstr_sig_map;
      init_abstr_ref:=false
      )
    else 
      () (* keep old sig abstraction*)
    );
(* abstr clauses *)
    List.iter (apply_sig_abstr_to_clause ar_sig_env ar_cl_env) cl_list;
    ar_env

  

(*-----------------*)
let ps_loop_time_limit ~start_time ~time_limit =
  match time_limit with
  | Undef -> Undef
  | Def time_limit ->
      let used_time = (Unix.gettimeofday ()) -. start_time in
      let ps_loop_time_limit =  time_limit -. used_time in
      Def(ps_loop_time_limit) (* can be negative *)

  let clear_solver_assumptions ar_env = (* used after solving is finished *)
    Prop_solver_exchange.rm_sim_adjoint_lits ar_env.ar_adjoint_preds; 
    Prop_solver_exchange.remove_solver_assumptions (TSet.elements (ar_env_get_all_assumptions ar_env))


(*--------main loop---------*)

let rec abstr_ref_gr_filter_loop ~time_limit input_clauses =
(*  !current_options.resolution_flag <- false; *)
(*  !current_options.inst_prop_sim_given <- false;
  !current_options.inst_prop_sim_new <- false;
  !current_options.res_prop_simpl_given <-false;
  !current_options.res_prop_simpl_new <- false;
*)
  let start_time = Unix.gettimeofday () in
  let ar_env = init_ar_env input_clauses in

  while true 
  do
    try 
      Statistics.incr_int_stat 1 Statistics.abstr_arg_filter_cycles;
      glb_abstr_loop_cnt := !glb_abstr_loop_cnt + 1;
      dbg D_trace (lazy ("abstr_ref_gr_filter_loop: "^(string_of_int !glb_abstr_loop_cnt)));      
(* adjoint pred *)
      let adjoint_pred = create_adjoint_pred !glb_abstr_loop_cnt in
      ar_env.ar_adjoint_preds <- [add_neg_atom adjoint_pred]; (* one adjoint predicate *)      

(* add adjoint to assumptions*)
(*      ar_env.ar_current_assumptions <- TSet.add adjoint_pred ar_env.ar_current_assumptions;      *)
      
      let ar_all_clauses = ar_env_get_all_clauses ar_env in 
      let ar_all_assumptions = ar_env_get_all_assumptions ar_env in

      Prop_solver_exchange.add_sim_adjoint_lits ar_env.ar_adjoint_preds; (* currently irrelevant since ~only_norm:true  when adding assumptions *)

(* for now we do not use assumptions for simplifications *)
      Prop_solver_exchange.add_solver_assumptions ~only_norm:true  (TSet.elements ar_all_assumptions);
      Prop_solver_exchange.clear_soft_assumptions ();
      (if (Prop_solver_exchange.solve ()) = PropSolver.Unsat
      then
	raise Unsatisfiable_gr
      );

      let ps_time_limit = ps_loop_time_limit ~start_time ~time_limit in
      Proof_search_loop.ps_full_loop ~time_limit:ps_time_limit (CSet.elements ar_all_clauses);                     
    with 
    | Unsatisfiable_gr_na -> 
        dbg D_trace (lazy ("Unsatisfiable_gr_na"));
        clear_solver_assumptions ar_env;
        raise Unsatisfiable_gr_na
    
    | Unsatisfiable_gr ->
        assert (Prop_solver_exchange.soft_assumptions_is_empty ());	 
	dbg D_trace (lazy ("Unsatisfiable_gr"));
        let uc = Prop_solver_exchange.get_unsat_core ~soft:false () in
	let uc_assumptions = UnsatCore.get_assumptions uc in 
        dbg D_uc (lazy ("uc assumptions: "^(Term.term_list_to_string uc_assumptions)));
        let swp_assumptions = TSet.inter (TSet.of_list uc_assumptions) (ar_env_get_all_assumptions ar_env) in
        let swp_assumptions_list = TSet.elements swp_assumptions in
        dbg D_uc (lazy ("UC:SWP:ASS: "^(Term.term_list_to_string swp_assumptions_list)));
        dbg D_uc (lazy ("UC:CL: "^(Clause.clause_list_to_string ( UnsatCore.get_clauses uc))));
        
        clear_solver_assumptions ar_env;
        if (TSet.is_empty swp_assumptions) (* trivial refinement *)
        then 
          begin
           (if (Prop_solver_exchange.solve ()) = PropSolver.Unsat
           then
	     raise Unsatisfiable_gr
           else
             failwith "abstr_ref: ground_filter_abstr_loop should be unsat"
           );           
          end            
        else 
          begin
            refine_ar_env ar_env swp_assumptions_list;
            (if !current_options.abstr_ref_until_sat
            then 
              begin
                let uc_orig_clauses = get_orig_clauses_swps ar_env swp_assumptions_list in
                let new_time_limit = ps_loop_time_limit ~start_time ~time_limit in
                refine_uc_until_sat ~time_limit:new_time_limit uc_orig_clauses;
              end
            else
              ()
            );
          end
    | Proof_search_loop.PS_loop_time_out(full_loop_counter) -> 
        clear_solver_assumptions ar_env;
        raise (Proof_search_loop.PS_loop_time_out(full_loop_counter))
    |x -> 
        clear_solver_assumptions ar_env;
        raise x
  done
(*--------------- only for EPR ? ---*)
and refine_uc_until_sat ~time_limit clauses =
  dbg D_uc_ref (lazy ("refine_uc_until_sat"));
  assert glb_reuse_sig_abstr_flag; (* assume abstr shared *)
  try 
    abstr_ref_gr_filter_loop ~time_limit clauses
  with 
  | Instantiation_loop.Inst_satisfiable _ | Resolution_loop.Res_satisfiable _ -> 
      (
       dbg D_uc_ref (lazy ("refine_uc_until_sat: SAT"));
      )

(*
      if (sym == Symbol.symb_typed_equality)
      then 
	let (type_term_eq, t, s) = get_triple_from_list (Term.arg_to_list args) in
	Def(Eq_type_term (type_term_eq, t, s))
      else
	Undef
*)

  

(* let symb_to_filter *)

(* 0 - value *)
(* p[f(t,x,s),u] -> p([f_0_1_2(t,s)]) *)
(* axiom: f_0_1_2(u,v) <-> f(u,y,v)   *)

(* we don't need to guard such axs  *)
(* guard only replacements *)
(*    
let abstr_refine uc_clauses 

type ar_over_env = 
    {
     ar_abstr_to_
   }
*)
