(*----------------------------------------------------------------------(C)-*)
(*   This file is part of iProver - a theorem prover for first-order logic. *)
(*   see the LICENSE  file for the license                                  *)

open Lib
open Options
open Logic_interface

(*----- debug modifiable part-----*)
let dbg_flag = false

type dbg_gr =
  |D_trace
  |D_uc
  |D_uc_ref

let dbg_gr_to_str = function
  |D_trace -> "trace"
  |D_uc -> "uc"
  |D_uc_ref -> "uc_ref"

let dbg_groups =
  [
    D_uc;
    D_uc_ref;
  ]

let module_name = "axiom_selection"

(*----- debug fixed part --------*)
let () = dbg_flag_msg dbg_flag module_name

let dbg group str_lazy =
  dbg_out_pref dbg_flag dbg_groups group dbg_gr_to_str module_name str_lazy

let dbg_env group f =
  dbg_env_set dbg_flag dbg_groups group f
(*----- debug -----*)

type symb_filter = symbol * int

module SF_Key =
struct
  type t = symb_filter
  let compare (f1,args1) (f2,args2) =
    if (f1 == f2)
    then
      Pervasives.compare args1 args2
    else
      Symbol.compare f1 f2
end

module SF_Map = Map.Make(SF_Key)
module SF_Set = Set.Make(SF_Key)
module TLMap = Term.MapList

type abstr_cl_wrap = (* aclw *)
  {
    aclw_orig_cl : clause;
    aclw_lits : lit list;    (* without the switch pred *)
    aclw_switch_pred : term; (* switch_pred -- prop pred *)
    aclw_acl : clause;       (* acl: aclw_lits \/ ~switch_pred *)
  }

type abstr_sig = (* asg; global abstr symbols *)
  {
    mutable asg_sf_to_as : symbol SF_Map.t;    (* (f,no_args) -> abstr_symb *)
    mutable asg_as_to_sf : SF_Set.t SMap.t; (* abstr_symb -> (f,no_args) *)
  }

type as_sig_env =  (* current abstr_sig env *)
  {
    (* global *)
    mutable ase_abstr_sig     : abstr_sig;
    (* current as *)
    mutable ase_symb_to_asymb : symbol SMap.t; (* symb -> abstr_symb *)
  }

type abstr_clauses = (* global sotrage *)
  {
    (* switch_pred -> abstr_cl_wrap  *)
    mutable as_swp_to_aclw       : abstr_cl_wrap TMap.t;
    (* lists of abstr literals -> switch preds *)
    mutable as_abstr_lits_to_swp : term TLMap.t;
  }

type as_cl_env = (* as environment containing current abstr of clauses *)
  {
    (* global *)
    mutable as_abstr_clauses   : abstr_clauses;
    mutable as_acl_set         : CSet.t;
    mutable as_concretised_set : CSet.t;
    mutable as_cl_assumptions  : TSet.t;
  }

type as_env =
  {
    mutable as_orig_clause_set : CSet.t;
    mutable as_sig_env         : as_sig_env;
    mutable as_cl_env          : as_cl_env;
  }

(* counter should be global since after schedule restarts the counter should not
   be reset *)
let glb_swp_cnt = ref 0

let create_switch_pred cnt =
  let sw_pred_name = "$$iProver_as_switch_pred_"^(string_of_int cnt) in
  add_pred_0 ~is_special:true sw_pred_name

let ps_loop_time_limit ~start_time ~time_limit =
  match time_limit with
  | Undef -> Undef
  | Def time_limit ->
    let used_time = (Unix.gettimeofday ()) -. start_time in
    let ps_loop_time_limit =  time_limit -. used_time in
    Def(ps_loop_time_limit) (* can be negative *)

let num_of_args args = List.length (Term.arg_to_list args)

let sf_to_str sf =
  let (sym, nargs) = sf in
  (Symbol.to_string sym) ^ ":" ^ (string_of_int nargs)

let sf_list_to_str sf_lst =
  let sf_strs = List.map sf_to_str sf_lst in
  String.concat "," sf_strs

let sf_set_to_str sf_set =
  let sf_lst = SF_Set.elements sf_set in
  sf_list_to_str sf_lst

let smap_to_str e_to_str map =
  let bindings = SMap.bindings map in
  let f (k,e) =
    (Symbol.to_string k) ^ " -> " ^ (e_to_str e)
  in
  String.concat "\n" (List.map f bindings)

let create_abstr_sig () =
  {
    asg_sf_to_as = SF_Map.empty;
    asg_as_to_sf = SMap.empty;
  }

let create_as_sig_env () =
  {
    ase_abstr_sig     = create_abstr_sig ();
    ase_symb_to_asymb = SMap.empty;
  }

let create_abstr_clauses () =
  {
    as_swp_to_aclw        = TMap.empty;
    as_abstr_lits_to_swp  = TLMap.empty;
  }

(* global; reused after restarts *)
let glb_abstr_sig = create_abstr_sig ()

(* global reuse of glb_abstr_sig *)
let create_as_sig_env_glb () =
  let as_sig_env = create_as_sig_env () in
  as_sig_env.ase_abstr_sig <- glb_abstr_sig;
  as_sig_env

let create_as_cl_env () =
  {
    as_abstr_clauses = create_abstr_clauses ();
    as_acl_set = CSet.empty;
    as_concretised_set = CSet.empty;
    as_cl_assumptions = TSet.empty;
  }

let glb_abstr_clauses = create_abstr_clauses ()

let create_as_cl_env_glb () =
  let as_cl_env = create_as_cl_env () in
  as_cl_env.as_abstr_clauses <- glb_abstr_clauses;
  as_cl_env

let as_cl_env_clear as_cl_env = (* keep only global as_abstr_clauses *)
  as_cl_env.as_acl_set <- CSet.empty;
  as_cl_env.as_concretised_set <- CSet.empty;
  as_cl_env.as_cl_assumptions <- TSet.empty

(* keep current sig abstraction globally *)
let glb_as_sig_env = create_as_sig_env_glb ()

(* global reuse of as_sig_env/as_cl_env *)
let create_as_env_glb () =
  {
    as_orig_clause_set = CSet.empty;
    as_sig_env         = glb_as_sig_env; (* create_as_sig_env_glb (); *)
    as_cl_env          = create_as_cl_env_glb ();
  }

let add_as_to_sf_set abstr_symb symb_filter abstr_sig =
  try
    let sf_set = SMap.find abstr_symb abstr_sig.asg_as_to_sf in
    let new_sf_set = SF_Set.union sf_set (SF_Set.singleton symb_filter) in
    abstr_sig.asg_as_to_sf <-
      SMap.add abstr_symb new_sf_set abstr_sig.asg_as_to_sf
  with
  | Not_found ->
    let new_sf_set = SF_Set.singleton symb_filter in
    abstr_sig.asg_as_to_sf <-
      SMap.add abstr_symb new_sf_set abstr_sig.asg_as_to_sf

let get_cl_switch_pred abstr_clauses lit_list =
  dbg D_trace (lazy ("get_cl_switch_pred:\n\t"^
                           (Term.term_list_to_string lit_list)));
  let lit_list = Term.remove_duplicate_terms lit_list in
  try
    TLMap.find lit_list abstr_clauses.as_abstr_lits_to_swp
  with
  | Not_found ->
    let sw_pred = create_switch_pred !glb_swp_cnt in
    dbg D_trace (lazy ("get_cl_switch_pred:\n\tNot_found: "^
                             (Term.term_list_to_string lit_list)^
                             "\n\tswp: "^(Term.to_string sw_pred)));
    abstr_clauses.as_abstr_lits_to_swp <-
      TLMap.add lit_list sw_pred abstr_clauses.as_abstr_lits_to_swp;
    glb_swp_cnt := !glb_swp_cnt + 1;
    sw_pred

let abstr_sig_add_abstr_symb abstr_sig orig_symb nargs =
  let symb_filter = (orig_symb, nargs) in
  let abstr_symb =
    try
      let abstr_symb = SF_Map.find symb_filter abstr_sig.asg_sf_to_as in
      abstr_symb
    with
    | Not_found ->
      let (arg_types, val_type) = Symbol.get_stype_args_val_def orig_symb in
      let arg_types_str =
        String.concat "_" (List.map Symbol.to_string arg_types) in
      let name = String.concat "_" [
          "axm_sel"; string_of_int nargs;
          (Symbol.to_string val_type); arg_types_str
        ] in
      let stype = Symbol.get_type orig_symb in
      let abstr_symb = create_symbol name stype in
      (* add abstr_symb to env. *)
      abstr_sig.asg_sf_to_as <-
        SF_Map.add symb_filter abstr_symb abstr_sig.asg_sf_to_as;
      add_as_to_sf_set abstr_symb symb_filter abstr_sig;
      abstr_symb
  in
  dbg D_trace (lazy ("abstr_sig_add_abstr_symb: orig: " ^
                           (Symbol.to_string orig_symb) ^ " -> " ^
                           "abstr: " ^ (Symbol.to_string abstr_symb)
                          ));
  abstr_symb

(* lit_list wihtout switch pred; returns (abstr_clause, swp) *)
let abstr_clauses_add abstr_clauses orig_clause abstr_lits =
  (* normalise ? or is it ok because orig_lits are normalised ?*)
  let switch_pred = get_cl_switch_pred abstr_clauses abstr_lits in
  try
    let abstr_cl_warp = TMap.find switch_pred abstr_clauses.as_swp_to_aclw in
    (abstr_cl_warp.aclw_acl, switch_pred)
  with
    Not_found ->
    begin
      let compl_switch_pred = add_neg_atom switch_pred in
      (* filtered clause *)
      let tstp_source = Clause.tstp_source_arg_filter_abstr orig_clause in
      let abstr_clause =
        create_clause tstp_source (compl_switch_pred::abstr_lits) in
      Clause.inherit_conjecture abstr_clause orig_clause;
      let abstr_cl_wrap = (* aclw *)
        {
          aclw_orig_cl = orig_clause;
          aclw_lits = abstr_lits;         (* without the switch pred *)
          aclw_switch_pred = switch_pred; (* switch_pred -- prop pred *)
          aclw_acl = abstr_clause;        (* acl: ~aclw_lits \/ switch_pred *)
        }
      in
      dbg D_trace (lazy ("abstr_clauses_add abstr_clause: "^
                               (Clause.to_string abstr_clause)));

      abstr_clauses.as_swp_to_aclw <-
        TMap.add switch_pred abstr_cl_wrap abstr_clauses.as_swp_to_aclw;
      (abstr_clause, switch_pred)
    end

(* remove abstraction of the orig_symb from the current abstraction *)
let as_sig_env_rm_abstr_symb as_sig_env orig_symb =
  as_sig_env.ase_symb_to_asymb <-
    SMap.remove orig_symb as_sig_env.ase_symb_to_asymb; ()

let as_sig_env_replace_abstr_symb as_sig_env orig_symb nargs =
  let abstr_sig = as_sig_env.ase_abstr_sig in
  (* first remove current abstraction *)
  as_sig_env_rm_abstr_symb as_sig_env orig_symb;
  let abstr_symb = abstr_sig_add_abstr_symb abstr_sig orig_symb nargs in
  as_sig_env.ase_symb_to_asymb <-
    SMap.add orig_symb abstr_symb as_sig_env.ase_symb_to_asymb;
  abstr_symb

(* adds new abstr; old abstr is not removed ! use rm above *)
let as_cl_env_add_acl as_cl_env orig_clause abstr_lits =
  let as_abstr_clauses = as_cl_env.as_abstr_clauses in
  let (abstr_cl, switch_pred) =
    abstr_clauses_add as_abstr_clauses orig_clause abstr_lits in
  as_cl_env.as_acl_set <- CSet.add abstr_cl as_cl_env.as_acl_set;
  as_cl_env.as_cl_assumptions <-
    TSet.add switch_pred as_cl_env.as_cl_assumptions

(* returns (abstr_symb, nargs) or raises Not_found  *)
let get_abstr_symb as_sig_env orig_symb =
  let abstr_sig = as_sig_env.ase_abstr_sig in
  let abstr_symb  = SMap.find orig_symb as_sig_env.ase_symb_to_asymb in
  (* let (_, nargs) = SMap.find abstr_symb abstr_sig.asg_as_to_sf in *)
  let sf_set = SMap.find abstr_symb abstr_sig.asg_as_to_sf in
  let (_, nargs) = SF_Set.min_elt sf_set in
  (abstr_symb, nargs)

let as_sig_env_all_concretised as_sig_env =
  SMap.is_empty as_sig_env.ase_symb_to_asymb

let rec abstr_term as_sig_env t =
  match t with
  |Term.Fun (orig_symb, args, _info) ->
    let args_list = Term.arg_to_list args in
    let (new_symb, new_args) =
      try
        (*can raise Not_found *)
        let (new_symb, nargs) = get_abstr_symb as_sig_env orig_symb in
        let new_args =  List.map (abstr_term as_sig_env) args_list in
        (new_symb, new_args)
      with
      | Not_found ->
        let new_args = List.map (abstr_term as_sig_env) args_list in
        (orig_symb, new_args)
    in
    add_fun_term new_symb new_args
  | Term.Var _ -> t

let as_cl_env_add_concr_cl as_cl_env concrete_cl =
  as_cl_env.as_concretised_set <-
    CSet.add concrete_cl as_cl_env.as_concretised_set

(* abstracts orig_clause *)
(* changes current clauses in as_cl_env *)
let apply_sig_abstr_to_clause as_sig_env as_cl_env orig_clause =
  dbg D_trace (lazy ("apply_sig_abstr_to_clause: orig_clause: "^
                           (Clause.to_string orig_clause)));
  let orig_lits = Clause.get_literals orig_clause in
  let abstr_lit lit =
    add_term_db (Term.apply_to_atom (abstr_term as_sig_env) lit) in
  let abstr_lits = List.map abstr_lit orig_lits in
  let non_trivial =
    List.exists2 (fun l1 l2 -> not (l1 == l2)) orig_lits abstr_lits in
  if non_trivial
  then
    begin (* add abstr cl *)
      as_cl_env_add_acl as_cl_env orig_clause abstr_lits
    end
  else
    begin (* orig_clause = abstr *)
      as_cl_env_add_concr_cl as_cl_env orig_clause
    end

(* returns: symb_to_sf is a map symb -> currenft arg filter for this symbol *)
let rec abstr_tsig symb_to_sf t =
  match t with
  | Term.Fun(symb,args,_) when (Symbol.is_pred symb || Symbol.is_fun symb) &&
                               not (Term.is_const_term t) ->
    let args_list = Term.arg_to_list args in
    let nargs = List.length args_list in
    let new_symb_to_sf = SMap.add symb nargs symb_to_sf in
    List.fold_left abstr_tsig new_symb_to_sf args_list
  |_-> symb_to_sf

let abstr_cl_list_sig cl_list =
  let abstr_clsig symb_to_sf c =
    let lits = Clause.get_literals c in
    let atoms = List.map Term.get_atom lits in
    List.fold_left abstr_tsig symb_to_sf atoms
  in
  List.fold_left abstr_clsig SMap.empty cl_list

let init_abstr_ref = ref true

let init_as_env cl_list =
  let as_env = create_as_env_glb () in
  let as_sig_env = as_env.as_sig_env in
  let as_cl_env = as_env.as_cl_env in
  as_env.as_orig_clause_set <- CSet.of_list cl_list;
  (* fill signature filter *)
  ( if !init_abstr_ref then
      let abstr_sig_map = abstr_cl_list_sig cl_list in
      SMap.iter
        (fun orig_symb nargs ->
           ignore (as_sig_env_replace_abstr_symb as_sig_env orig_symb nargs)
        ) abstr_sig_map;
      init_abstr_ref := false
    else
      () (* keep old sig abstraction*)
  );
  (* abstr clauses *)
  List.iter (apply_sig_abstr_to_clause as_sig_env as_cl_env) cl_list;
  as_env

let as_cl_env_get_all_clauses as_cl_env =
  dbg D_trace (lazy ("-----------------------"));
  dbg D_trace (lazy ("as_cl_env_get_all_clauses: abstr clauses:\n"^
                           (Clause.clause_list_to_string
                              (CSet.elements as_cl_env.as_acl_set))));
  dbg D_trace (lazy ("-----------------------"));
  dbg D_trace (lazy ("as_cl_env_get_all_clauses: concretised clauses:\n"^
                           (Clause.clause_list_to_string
                              (CSet.elements as_cl_env.as_concretised_set))));
  CSet.union as_cl_env.as_acl_set as_cl_env.as_concretised_set

let as_cl_env_get_all_assumptions as_cl_env = as_cl_env.as_cl_assumptions

let as_env_get_all_clauses as_env = as_cl_env_get_all_clauses as_env.as_cl_env

let as_env_get_all_assumptions as_env =
  as_cl_env_get_all_assumptions as_env.as_cl_env

let clear_solver_assumptions as_env = (* used after solving is finished *)
  let assumptions = TSet.elements (as_env_get_all_assumptions as_env) in
  dbg D_trace (lazy ("\t*** cleaning assumptions ***"));
  Prop_solver_exchange.remove_solver_assumptions assumptions

(* returns (orig_symb, nargs) or raises Not_found  *)
let abstr_sig_get_orig_symb abstr_sig abstr_symb =
  SMap.find abstr_symb abstr_sig.asg_as_to_sf

let refine_symb_as_sig_env as_sig_env orig_symb =
  as_sig_env_rm_abstr_symb as_sig_env orig_symb  (* concretise this symbol *)

let refine_symbs_as_sig_env as_sig_env orig_symbs =
  SSet.iter (refine_symb_as_sig_env as_sig_env) orig_symbs

let get_orig_sig_to_refine as_env swps =
  let as_sig_env = as_env.as_sig_env in
  let as_cl_env = as_env.as_cl_env in
  let as_abstr_clauses = as_cl_env.as_abstr_clauses in
  let abstr_symbs_uc =
    let f asymb_rest swp =
      try
        let abstr_cl_wrap = TMap.find swp as_abstr_clauses.as_swp_to_aclw in
        let acl_asymbs = Clause.find_all_sym
            ~is_relevant_symb:(fun _ -> true) abstr_cl_wrap.aclw_acl in
        SSet.union acl_asymbs asymb_rest
      with
        Not_found -> asymb_rest
    in
    List.fold_left f SSet.empty swps
  in
  dbg D_trace (lazy ("get_orig_sig_to_refine: abstr_symbs_uc:\n" ^
                           (Symbol.list_to_string
                              (SSet.elements abstr_symbs_uc))));
  let orig_symbs_uc =
    let f abstr_symb symb_rest =
      try
        let sf_set =
          abstr_sig_get_orig_symb as_sig_env.ase_abstr_sig abstr_symb in
        let orig_symb_lst =
          List.map (fun (os, _) -> os) (SF_Set.elements sf_set) in
        let orig_symb_set = SSet.of_list orig_symb_lst in
        SSet.union orig_symb_set symb_rest
      with
        Not_found ->
        symb_rest
    in
    SSet.fold f abstr_symbs_uc SSet.empty
  in
  dbg D_trace (lazy ("get_orig_sig_to_refine: orig_symbs_uc:\n" ^
                           (Symbol.list_to_string
                              (SSet.elements orig_symbs_uc))));
  orig_symbs_uc

let get_orig_clauses_swps as_env swps =
  let as_cl_env = as_env.as_cl_env in
  let as_abstr_clauses = as_cl_env.as_abstr_clauses in
  let f cl_rest swp =
    try
      let abstr_cl_wrap = TMap.find swp as_abstr_clauses.as_swp_to_aclw in
      let orig_clause = abstr_cl_wrap.aclw_orig_cl in
      orig_clause::cl_rest
    with
      Not_found -> cl_rest
  in
  List.fold_left f [] swps

let refine_as_env as_env swps =  (* swps from uc *)
  dbg D_trace (lazy ("\t*** refine_as_env ***"));
  let as_sig_env = as_env.as_sig_env in
  let as_cl_env = as_env.as_cl_env in
  let orig_symbs = get_orig_sig_to_refine as_env swps in
  dbg D_trace (lazy ("refine_as_env: orig_symbs to refine:\n"^
                           (Symbol.list_to_string (SSet.elements orig_symbs))));
  dbg D_trace (lazy ("refine_as_env before refine: symb_to_asymb:\n"^
                           (smap_to_str Symbol.to_string
                              as_sig_env.ase_symb_to_asymb)));
  refine_symbs_as_sig_env as_sig_env orig_symbs;
  dbg D_trace (lazy ("refine_as_env after refine: symb_to_asymb:\n"^
                           (smap_to_str Symbol.to_string
                              as_sig_env.ase_symb_to_asymb)));
  (* TODO: clean! *)
  as_cl_env_clear as_cl_env;
  if (as_sig_env_all_concretised as_sig_env)
  then
    as_cl_env.as_concretised_set <- as_env.as_orig_clause_set
  else (
    let apply_sigabstr_to_cl = apply_sig_abstr_to_clause as_sig_env as_cl_env in
    let orig_cl_set = as_env.as_orig_clause_set in
    (dbg D_trace (lazy ("\t*** as_orig_clause_set ***\n" ^
                              Clause.clause_list_to_string
                                (CSet.elements orig_cl_set))));
    CSet.iter apply_sigabstr_to_cl orig_cl_set
  )

let rec axiom_selection_loop ~time_limit input_clauses =
  let start_time = Unix.gettimeofday () in
  let as_env = init_as_env input_clauses in
  (dbg D_trace (lazy ("\t*** Input Clauses ***\n" ^
                            Clause.clause_list_to_string input_clauses)));
  while true
  do
    try
      Statistics.incr_int_stat 1 Statistics.abstr_arg_filter_cycles;
      let as_all_clauses = as_env_get_all_clauses as_env in
      let all_clauses = CSet.elements as_all_clauses in
      let as_all_assumptions = as_env_get_all_assumptions as_env in
      Prop_solver_exchange.add_solver_assumptions
        ~only_norm:true  (TSet.elements as_all_assumptions);
      Prop_solver_exchange.clear_soft_assumptions ();
      let ps_time_limit = ps_loop_time_limit ~start_time ~time_limit in
      (dbg D_trace (lazy ("\t*** Abstraction/Refinement Clauses ***\n" ^
                                Clause.clause_list_to_string all_clauses)));
      Proof_search_loop.ps_full_loop ~time_limit:ps_time_limit all_clauses;
    with
    | Unsatisfiable_gr_na ->
      dbg D_trace (lazy ("Unsatisfiable_gr_na"));
      clear_solver_assumptions as_env;
      raise Unsatisfiable_gr_na
    | Unsatisfiable_gr ->
      assert (Prop_solver_exchange.soft_assumptions_is_empty ());
	    dbg D_trace (lazy ("Unsatisfiable_gr"));
      let uc = Prop_solver_exchange.get_unsat_core ~soft:false () in
	    let uc_assumptions = UnsatCore.get_assumptions uc in
      dbg D_uc (lazy ("\tuc assumptions:\n" ^
                            (Term.term_list_to_string uc_assumptions)));
      let uc_ass_set = TSet.of_list uc_assumptions in
      let all_ass_set = as_env_get_all_assumptions as_env in
      let swp_assumptions = TSet.inter uc_ass_set all_ass_set in
      let swp_assumptions_list = TSet.elements swp_assumptions in
      dbg D_uc (lazy ("\tUC:SWP:ASS:\n"^
                            (Term.term_list_to_string swp_assumptions_list)));
      dbg D_uc (lazy ("\tUC:CL:\n"^
                            (Clause.clause_list_to_string
                               (UnsatCore.get_clauses uc))));
      clear_solver_assumptions as_env;
      dbg D_trace (lazy ("\t*** symb_to_asymb:\n"^
                               (smap_to_str Symbol.to_string
                                  as_env.as_sig_env.ase_symb_to_asymb)));
      dbg D_trace (lazy ("\t*** as_to_sf:\n"^
                               (smap_to_str sf_set_to_str
                                  as_env.as_sig_env.ase_abstr_sig.asg_as_to_sf)
                              ));
      if not (TSet.is_empty swp_assumptions) (* trivial refinement *)
      then
        begin
          refine_as_env as_env swp_assumptions_list;
          if !current_options.abstr_ref_until_sat
          then (
            let uc_orig_clauses =
              get_orig_clauses_swps as_env swp_assumptions_list in
            let new_time_limit = ps_loop_time_limit ~start_time ~time_limit in
            refine_uc_until_sat ~time_limit:new_time_limit uc_orig_clauses;
          )
        end
    | Proof_search_loop.PS_loop_time_out(full_loop_counter) ->
      clear_solver_assumptions as_env;
      raise (Proof_search_loop.PS_loop_time_out(full_loop_counter))
    |x ->
      clear_solver_assumptions as_env;
      raise x
  done
(*--------------- only for EPR ? ---*)
and refine_uc_until_sat ~time_limit clauses =
  dbg D_uc_ref (lazy ("refine_uc_until_sat"));
  try
    axiom_selection_loop ~time_limit clauses
  with
  | Instantiation_loop.Inst_satisfiable _ | Resolution_loop.Res_satisfiable _ ->
    (
      dbg D_uc_ref (lazy ("refine_uc_until_sat: SAT"));
    )
